/*
    Copyright (C) 2002  bjk <bjk@arbornet.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <pwd.h>
#include <syslog.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "bubblegum.h"

int is_file(const char *filename)
{
    struct stat mystat;

    if (stat(filename, &mystat) == -1)
	return errno;

    if (!(mystat.st_mode & S_IFREG))
	return -1;

    return 0;
}

void logit(const char *format, ...)
{
    FILE *logfp;
    va_list ap;
    char buf[LINE_MAX], line[LINE_MAX], timebuf[MAX_TIME_LEN];
    time_t now;
    struct tm *mytm;

    if (loglevel == 0 && usesyslog == 0)
	return;

    timebuf[0] = line[0] = buf[0] = 0;

    va_start(ap, format);

    if (usesyslog) {
	(void)openlog(PROGNAME, LOG_PID, SYSLOG_FACILITY);
	(void)vsyslog(SYSLOG_PRIORITY, format, ap);
	(void)closelog();

        va_end(ap);
	return;
    }

    vsnprintf(buf, sizeof(buf), format, ap);
    va_end(ap);

    if ((logfp = fopen(logfile, "a")) == NULL) {
	perror(logfile);
	return;
    }

    time(&now);
    mytm = localtime(&now);
    strftime(timebuf, sizeof(timebuf), TIMEFORMAT, mytm);
    timebuf[strlen(timebuf)] = '\0';

    snprintf(line, sizeof(line), "[%s] %i %s\n", timebuf, getpid(), buf);
    fprintf(logfp, "%s", line);

    fflush(logfp);
    fclose(logfp);
    return;
}

char *diffstr(char dest[], size_t size, time_t compare, time_t basetime)
{
    int diff = 0;
    int day, hour, min, sec, i;

    day = hour = min = sec = i = 0;
    diff = compare - basetime;

    day = diff / 86400;
    i = diff % 86400;

    hour = i / 3600;
    i = i % 3600;

    min = i / 60;
    i = i % 60;

    sec = i;

    snprintf(dest, size, "%id,%ih,%im,%is", day, hour, min, sec);
    return (char *)dest;
}

char *fullpath(const char *wd, char *filename, char dest[], size_t size)
{
    dest[0] = 0;

    if (filename[0] == '/')
	strncpy(dest, filename, size);
    else {
	strncpy(dest, wd, size);
	strncat(dest, "/", size);
	strncat(dest, filename, size);
    }

    return (char *)dest;
}

/* convert integer to ascii. this is by ndf@linux.mit.edu */
static char *itoa(int i)
{
    char buf[13];
    char *pos = buf + sizeof(buf) - 1;
    unsigned int u;
    int negetive = 0;

    if (i < 0) {
	negetive = 1;
	u = ((unsigned int)(-(1+i))) + 1;
    }
    else
	u = i;

    *pos = 0;

    do {
	*--pos = '0' + (u % 10);
	u /= 10;
    } while (u);

    if (negetive)
	*--pos = '-';

    return pos;
}

char *changelist(unsigned char changes)
{
    char buf[LINE_MAX], *s = NULL;

    buf[0] = 0;

    if (TEST_BIT(changes, ACCESS))
	s = strncat(buf, "ATIME,", sizeof(buf));

    if (TEST_BIT(changes, MODIFY))
	s = strncat(buf, "MTIME,", sizeof(buf));

    if (TEST_BIT(changes, CHANGE))
	s = strncat(buf, "CTIME,", sizeof(buf));

    if (TEST_BIT(changes, ERROR))
	s = strncat(buf, "ERROR,", sizeof(buf));

    if (!s)
	s = "DOH,";

    s[strlen(s) - 1] = 0;

    return (char *)s;
}

char *parsecmd(struct filelist *flist, char *line, time_t basetime)
{
    char buf[sizeof(command)];
    int i = 0;

    buf[0] = 0;

    if (line[0] == 0)
	return NULL;

    while (*line) {
        struct tm *mytm;
	char *s = NULL;
	char timebuf[MAX_TIME_LEN] = {0};
	int n;
	time_t blah = 0;

	if (*line == '%') {
	    switch (*++line) {
		case 'f':
		    s = (char *)flist->filename;
		    break;
		case 't':
		    /* kinda hokey. use the latest change time. */
		    if (flist->laststat.st_atime > blah)
			blah = flist->laststat.st_atime;

		    if (flist->laststat.st_mtime > blah)
			blah = flist->laststat.st_mtime;

		    if (flist->laststat.st_ctime > blah)
			blah = flist->laststat.st_ctime;

		    mytm = localtime(&blah);

		    strftime(timebuf, sizeof(timebuf), TIMEFORMAT, mytm);
		    timebuf[strlen(timebuf)] = '\0';
		    s = (char *)timebuf;
		    break;
		case 'b':
		    mytm = localtime(&basetime);
		    strftime(timebuf, sizeof(timebuf), TIMEFORMAT, mytm);
		    timebuf[strlen(timebuf)] = '\0';

		    s = (char *)timebuf;
		    break;
		case 'c':
		    s = changelist(changes);
		    break;
		case 'p':
		    s = itoa(getpid());
		    break;
		case 'e':
		    s = (flist->staterrno ? strerror(flist->staterrno) : " ");
		    break;
		case '%':
		    s = "%";
		    break;
		default:
		    break;
	    }
	}

	if (s) {
	    for (n = 0; n < strlen(s); n++) {
		if (strlen(buf) + 1 == sizeof(buf))
		    break;

		buf[i++] = s[n];
	    }

	    s = NULL;
	    n = *line++;
	    continue;
	}

	if (strlen(buf) + 1 == sizeof(buf))
	    break;

	buf[i++] = *line++;
    }

    buf[i] = 0;
    return (char *)strdup(buf);
}

void freelist(struct filelist *flisthead)
{
    struct filelist *cur, *next;

    for (cur = flisthead; cur; cur = next) {
	next = cur->next;
	free(cur->filename);
	free(cur);
    }

    return;
}

struct filelist *loadfile(const char *filename)
{
    FILE *fp;
    char *cp, buf[LINE_MAX];
    static int loaded;
    struct filelist *head, *flist;

    if ((fp = fopen(filename, "r")) == NULL)
	return NULL;

    if ((head = (struct filelist *)calloc(1, sizeof(struct filelist)))
	    == NULL) {
	if (loaded)
	    logit("calloc: %s", strerror(errno));
	else
	    perror("calloc");

	exit(errno);
    }

    flist = head;
    total = 0;

    while ((cp = fgets(buf, sizeof(buf), fp)) != NULL) {
	struct stat mystat;
	char tmp[sizeof(buf)];

	tmp[0] = buf[strlen(buf) - 1] = '\0';
	cp = fullpath(wd, cp, tmp, sizeof(tmp));

	if (stat(cp, &mystat) == -1) {
	    if (loaded)
		logit("%s: %s", cp, strerror(errno));
	    else
		perror(cp);

	    continue;
	}

	if ((flist->filename = (char *)calloc(1, strlen(cp) + 2)) == NULL) {
	    if (loaded)
		logit("calloc: %s", strerror(errno));
	    else
		perror("calloc");

	    exit(errno);
	}

	strncpy(flist->filename, cp, strlen(cp));
	flist->laststat = mystat;

	if ((flist->next = (struct filelist*) 
		    calloc(1, sizeof(struct filelist))) == NULL) {
	    if (loaded)
		logit("calloc: %s", strerror(errno));
	    else
		perror("calloc");

	    exit(errno);
	}

	flist = flist->next;
	total++;
    }

    flist->next = NULL;
    loaded = 1;

    fclose(fp);
    return head;
}

void catchsig(int signal)
{
    int status;

    switch (signal) {
	case SIGTERM:
	    logit("signal %i: terminating", signal);
	    exit(signal);
	    break;
	case SIGCHLD:
	    wait3(&status, WNOHANG, 0);
	    break;
	case SIGUSR1:
	    if (filenamelist[0]) {
	       	logit("signal %i: reloading %s", signal, filenamelist);
		reload = 1;
	    }
	    else
		logit("signal %i: no file specified (-f)", signal);

	    break;
	case SIGUSR2:
	    logit("signal %i: resetting changed bits", signal);
	    resetbits = 1;
	    break;
	default:
	    logit("signal %i: ignoring", signal);
	    break;
    }

    return;
}

/* test to see if any bits are set */
int checkbits(unsigned char blah)
{
    if (TEST_BIT(blah, ACCESS) ||
	    TEST_BIT(blah, MODIFY) ||
	    TEST_BIT(blah, CHANGE) ||
	    TEST_BIT(blah, ERROR))
	return 1;

    return 0;
}

void changelog(struct filelist *flist, time_t basetime)
{
    char *str = NULL, *tmp;
    char buf[255] = {0};

    if (TEST_BIT(changes, ACCESS) && TEST_BIT(check, ACCESS)) {
	str = "ATIME";
	tmp = diffstr(buf, sizeof(buf), flist->laststat.st_atime, basetime); 

	if (!TEST_BIT(flist->changed, ACCESS)) {
	    if (loglevel == 1) {
		if (!checkbits(flist->changed))
		    logit("%s \"%s\" (%s)", str, flist->filename, tmp); 
	    }
	    else
		logit("%s \"%s\" (%s)", str, flist->filename, tmp);
	}

	SET_BIT(flist->changed, ACCESS);
    }

    if (TEST_BIT(changes, MODIFY) && TEST_BIT(check, MODIFY)) {
	str = "MTIME";
	tmp = diffstr(buf, sizeof(buf), flist->laststat.st_mtime, basetime); 

	if (!TEST_BIT(flist->changed, MODIFY)) {
	    if (loglevel == 1) {
		if (!checkbits(flist->changed))
		    logit("%s \"%s\" (%s)", str, flist->filename, tmp); 
	    }
	    else
	        logit("%s \"%s\" (%s)", str, flist->filename, tmp); 
	}

	SET_BIT(flist->changed, MODIFY);
    }

    if (TEST_BIT(changes, CHANGE) && TEST_BIT(check, CHANGE)) {
	str = "CTIME";
	tmp = diffstr(buf, sizeof(buf), flist->laststat.st_ctime, basetime); 

	if (!TEST_BIT(flist->changed, CHANGE)) {
	    if (loglevel == 1) {
		if (!checkbits(flist->changed))
		    logit("%s \"%s\" (%s)", str, flist->filename, tmp); 
	    }
	    else
	        logit("%s \"%s\" (%s)", str, flist->filename, tmp); 
	}

	SET_BIT(flist->changed, CHANGE);
    }

    if (TEST_BIT(changes, ERROR) && TEST_BIT(check, ERROR)) {
	str = "ERROR";

	if (!TEST_BIT(flist->changed, ERROR)) {
	    if (loglevel == 1) {
		if (!checkbits(flist->changed))
		    logit("%s \"%s\" %s", str, flist->filename,
			    strerror(flist->staterrno)); 
	    }
	    else
		logit("%s \"%s\" %s", str, flist->filename,
		       	strerror(flist->staterrno)); 
	}

	SET_BIT(flist->changed, ERROR);
    }

    return;
}

void resetchanged(struct filelist *flist)
{
    RESET_BIT(flist->changed, ACCESS);
    RESET_BIT(flist->changed, MODIFY);
    RESET_BIT(flist->changed, CHANGE);
    RESET_BIT(flist->changed, ERROR);

    return;
}

void doit(struct filelist *flisthead)
{
    static int runs;

    logit("%s %s (pri: %i, lvl: %i, int: %is, files: %i)", PROGNAME, VERSION,
	   niceness, loglevel, seconds, total);

    while (1) {
	struct timeval tv;
        struct filelist *flist, *cur;
	int err, status;
	pid_t pid;

	if (reload) {
	    if ((err = is_file(filenamelist))) {
		if (err > 0)
		    logit("%s: %s", filenamelist, strerror(err));
		else
		    logit("%s: not a file", filenamelist);
	    }
	    else {
		freelist(flisthead);

		if ((flisthead = loadfile(filenamelist)) == NULL) { 
		    logit("%s", "error reloading file");
		    raise(SIGTERM);
		}
		else 
		    logit("%s: %i %s", filenamelist, total, 
			    (total > 1 ? "files" : "file"));
	    }

	    reload = runs = 0;
	}

	if (resetbits) {
	    for (cur = flisthead; cur; cur = flist) {
		flist = cur->next;
		resetchanged(cur);
	    }

	    resetbits = 0;
	}

	flist = flisthead;

	while (flist->next != NULL) {
	    char *cmd = NULL;
	    int err;
	    time_t basetime = 0;
	    struct stat mystat;

	    if (stat(flist->filename, &mystat) == -1) {
		if (flist->staterrno == errno) {
		    flist = flist->next;
		    continue;
		}

		SET_BIT(changes, ERROR);
		flist->staterrno = errno;
	    }

	    if (mystat.st_atime != flist->laststat.st_atime) {
		SET_BIT(changes, ACCESS);
		basetime = flist->laststat.st_atime;
	    }

	    if (mystat.st_mtime != flist->laststat.st_mtime) {
		SET_BIT(changes, MODIFY);
		basetime = flist->laststat.st_mtime;
	    }

	    if (mystat.st_ctime != flist->laststat.st_ctime) {
		SET_BIT(changes, CHANGE);
		basetime = flist->laststat.st_ctime;
	    }

	    if (checkbits(changes)) {
		flist->laststat = mystat;

		changelog(flist, basetime);

		if (((totalruns && runs < totalruns) || !cmdruns ||
			    flist->cmdruns < cmdruns) && command[0]) {
		    cmd = parsecmd(flist, command, basetime);
		    flist->cmdruns++;
		    runs++;

		    switch ((pid = fork())) {
			case 0:
			    switch (fork()) {
				case 0:
				    if ((err = system(cmd)) == -1) {
					logit("system(): %s", strerror(errno));
					exit(errno);
				    }

				    logit("command exited with exit status "
					    "%i", err);
				    exit(err);
				case -1:
				    logit("fork(): %s", strerror(errno));
				    exit(errno);
				default:
				    exit(EXIT_SUCCESS);
			    }

			    _exit(EXIT_SUCCESS);
			case -1:
			    logit("fork(): %s", strerror(errno));
			    break;
			default:
			    wait(&status);

			    logit("fork #%i (pid: %i): %s", (totalruns ? runs :
					flist->cmdruns), pid, cmd);
			    break;
		    }

		    if (cmd)
			free(cmd);
		}

		if (loglevel == 3) 
		    resetchanged(flist);
	    }

	    RESET_BIT(changes, ACCESS);
	    RESET_BIT(changes, MODIFY);
	    RESET_BIT(changes, CHANGE);
	    RESET_BIT(changes, ERROR);

	    flist = flist->next;
	}

	tv.tv_sec = seconds;
	tv.tv_usec = 0;

	select(0, 0, 0, 0, &tv);
    }

    return;
}

static void usage(const char *pn)
{
    fprintf(stderr, "Usage: %s [-hvs] [-i secs] [-l <filename>] [-level] "
	    "[-e <command>]\n\t[-r runs] [-t runs] [-n niceness] -amc "
	    "-f <filename> | file ...\n", pn);
    fprintf(stderr, "  -a\tCheck the access time of the file.\n");
    fprintf(stderr, "  -m\tCheck the modification time of the file.\n");
    fprintf(stderr, "  -c\tCheck the change time of the inode.\n");
    fprintf(stderr, "  -e\tCommand to run when a file changes (see below).\n");
    fprintf(stderr, "  -r\tNumber of times to run command per file "
	    "(0=infinite). The default is %i.\n", CMDRUNS);
    fprintf(stderr, "  -t\tTotal number of times to run command (overrides "
	    "-r).\n");
    fprintf(stderr, "  -n\tNiceness (-20 through 20) of the daemon and command. "
	"The default is %i.\n", niceness);
    fprintf(stderr, "  -i\tInterval in seconds to check files. The default is "
	    "%i.\n", INTERVAL);
    fprintf(stderr, "  -f\tA file containing a list of files, one per line, "
	    "to watch.\n");
    fprintf(stderr, "  -l\tLog to an alternate file. The default is "
	    "~/%s.\n", LOGFILE);
    fprintf(stderr, "  -#\tLog level: 1=first change, 2=first of any "
	    "change, 3=all. Default is %i.\n", LOGLEVEL);
    fprintf(stderr, "  -s\tDisable logging to file (-l) and log to "
	    "syslogd(8).\n");
    fprintf(stderr, "  -h\tThis help text.\n");
    fprintf(stderr, "  -v\tVersion and copyright information.\n\n");
    fprintf(stderr, "The following expansions are available in a command "
	    "(remember to quote!):\n");
    fprintf(stderr, "  %%f - filename\t");
    fprintf(stderr, "\t%%t - time of (latest) change\t");
    fprintf(stderr, "%%b - base timestamp\n");
    fprintf(stderr, "  %%e - error info\t");
    fprintf(stderr, "%%%% - regular %% character\t");
    fprintf(stderr, "%%p - daemon PID\n");
    fprintf(stderr, "  %%c - change type(s): ATIME, MTIME, CTIME, "
	    "ERROR\n\n");
    fprintf(stderr, "Example: %s -i30 -amc -e 'echo \"%%t:%%p:%%c:%%f\" "
	    "| mail user' -f filelist\n", pn);

    exit(EXIT_FAILURE);
}

int main(int argc, char *argv[])
{
    int opt;
    char *pn;
    static struct sigaction sa;
    struct filelist *flisthead, *flist;
    struct passwd *passwd;

    pn = PROGNAME;

    if (getcwd(wd, sizeof(wd) - 1) == NULL) {
	perror("getcwd()");
	exit(errno);
    }

    wd[strlen(wd)] = '\0';

    if ((passwd = getpwuid(getuid())) == NULL) {
	fprintf(stderr, "getpwuid(): unknown error\n");
	exit(errno);
    }

    snprintf(logfile, sizeof(logfile), "%s/%s", passwd->pw_dir, LOGFILE);

    seconds = INTERVAL;
    SET_BIT(check, ERROR);
    loglevel = LOGLEVEL;
    cmdruns = CMDRUNS;

    errno = 0;

    if ((niceness = getpriority(PRIO_PROCESS, getpid())) == -1) {
	if (errno) {
	    perror("getpriority()");
	    exit(errno);
	}
    }

    while ((opt = getopt(argc, argv, "n:t:shvr:ae:i:mf:l:c123")) != -1) {
	int i, err;
	char tmp[2];

	switch (opt) {
	    case 'n':
		niceness = atoi(optarg);

		if (niceness > 20 || niceness < -20)
		    usage(pn);
		break;
	    case 's':
		usesyslog = 1;
		break;
	    case 'e':
		if (optarg[0] == 0)
		    usage(pn);

		strncpy(command, optarg, sizeof(command));
		break;
	    case 'l':
    	        fullpath(wd, optarg, logfile, sizeof(logfile));
		break;
	    case '1':
	    case '2':
	    case '3':
		snprintf(tmp, sizeof(tmp), "%c", opt);
		loglevel = atoi(tmp);
		break;
	    case 'f':
    	        fullpath(wd, optarg, filenamelist, sizeof(filenamelist));

		if ((err = is_file(filenamelist))) {
		    if (err > 0) {
			perror(filenamelist);
			exit(errno);
		    }
		    else
		       	fprintf(stderr, "%s: not a file\n", filenamelist);

		    exit(EXIT_FAILURE);
		}
		break;
	    case 't':
		for (i = 0; i < strlen(optarg); i++) {
		    if (isdigit(optarg[i]) == 0)
			usage(pn);
		}

		totalruns = atoi(optarg);
		break;
	    case 'r':
		for (i = 0; i < strlen(optarg); i++) {
		    if (isdigit(optarg[i]) == 0)
			usage(pn);
		}

		cmdruns = atoi(optarg);
		break;
	    case 'i':
		for (i = 0; i < strlen(optarg); i++) {
		    if (isdigit(optarg[i]) == 0)
			usage(pn);
		}

		seconds = atol(optarg);

		if (seconds < 1)
		    usage(pn);
		break;
	    case 'h':
		usage(pn);
		break;
	    case 'v':
		fprintf(stdout, "%s %s\n%s\n", PROGNAME, VERSION, COPYRIGHT);
		exit(EXIT_SUCCESS);
	    case 'a':
		SET_BIT(check, ACCESS);
		break;
	    case 'c':
		SET_BIT(check, CHANGE);
		break;
	    case 'm':
		SET_BIT(check, MODIFY);
		break;
	    default:
		usage(pn);
		break;
	}
    }

    if (!TEST_BIT(check, ACCESS) && !TEST_BIT(check, ACCESS) && 
	    !TEST_BIT(check, ACCESS))
	usage(pn);

    if (argc == optind && !filenamelist[0])
	usage(pn);

    if (filenamelist[0]) {
	if ((flisthead = loadfile(filenamelist)) == NULL) {
	    perror(filenamelist);
	    exit(errno);
	}
    }
    else {
	if ((flisthead = (struct filelist *)calloc(1, sizeof(struct filelist)))
		== NULL) {
	    perror("calloc");
	    exit(errno);
	}

	flist = flisthead;

	while (optind < argc) {
	    struct stat mystat;
	    char tmp[FILENAME_MAX], *cp;

	    tmp[0] = 0;

	    if (stat(argv[optind], &mystat) == -1) {
		perror(argv[optind++]);
		continue;
	    }

	    cp = fullpath(wd, argv[optind++], tmp, sizeof(tmp));

	    if ((flist->filename = (char *)calloc(1, strlen(cp) + 2)) == NULL) {
		perror("calloc");
		exit(errno);
	    }

	    strncpy(flist->filename, cp, strlen(cp));
	    flist->laststat = mystat;

	    if ((flist->next = (struct filelist*) 
			calloc(1, sizeof(struct filelist))) == NULL) {
		perror("calloc");
		exit(errno);
	    }

	    flist = flist->next;
	    total++;
	}

	flist->next = NULL;
    }

    if (!flisthead->filename) {
	fprintf(stderr, "No files to watch. Exiting.\n");
	exit(EXIT_FAILURE);
    }

    sigfillset(&sa.sa_mask);
    sigdelset(&sa.sa_mask, SIGTERM);
    sigdelset(&sa.sa_mask, SIGCHLD);
    sigdelset(&sa.sa_mask, SIGUSR1);
    sigdelset(&sa.sa_mask, SIGUSR2);

    if (sigprocmask(SIG_BLOCK, &sa.sa_mask, 0) == -1) {
	perror("sigprocmask()");
	exit(errno);
    }

    sa.sa_handler = catchsig;

    if (sigaction(SIGTERM, &sa, NULL) == -1) {
	perror("sigaction()");
	exit(errno);
    }

    if (sigaction(SIGCHLD, &sa, NULL) == -1) {
	perror("sigaction()");
	exit(errno);
    }

    if (sigaction(SIGUSR1, &sa, NULL) == -1) {
	perror("sigaction()");
	exit(errno);
    }

    if (sigaction(SIGUSR2, &sa, NULL) == -1) {
	perror("sigaction()");
	exit(errno);
    }

    if (setpriority(PRIO_PROCESS, getpid(), niceness) == -1) {
	perror("setpriority()");
	exit(errno);
    }

    switch (fork()) {
	case 0:
	    if (chdir("/") == -1) {
		perror("/");
		exit(errno);
	    }

	    if (setsid() == -1) {
		perror("setsid()");
		exit(errno);
	    }

	    freopen("/dev/null", "r", stdin);
	    freopen("/dev/null", "w", stdout);
	    freopen("/dev/null", "w", stderr);

	    doit(flisthead);
	    break;
	case -1:
	    perror("fork()");
	    exit(errno);
	    break;
	default:
	    exit(EXIT_SUCCESS);
	    break;
    }

    exit(EXIT_SUCCESS);
}
